struct tree
{
  struct tree *leftChild;
  struct tree *rightChild;
  int nObs, firstObsInInterval, maxPolyOrder, optPolyOrder;
  double **polyCoeffs, *likelihoods;
  double intervalLeft, intervalRight,optLikelihood;
};

void printTreeNode(struct tree *T, FILE *outfile)
{
  int i,k;
  fprintf(outfile, "Interval [%f,%f], polynomial order = %d\n",
	 T->intervalLeft,T->intervalRight, T->optPolyOrder);
  fprintf(outfile, "\tChebyshev polynomial coefficients:\t");
  for (k = 0; k < T->optPolyOrder; k++) {
    fprintf(outfile, "%lf\t",T->polyCoeffs[T->optPolyOrder-1][k]);
  }
  fprintf(outfile,"\n");
}

void printTree(struct tree *T, FILE *outFile)
{
  if (T == NULL) 
    return;
  if (T->optPolyOrder > -1)
    printTreeNode(T,outFile);
  else {
    printTree(T->leftChild,outFile);
    printTree(T->rightChild,outFile);
  }
}

void printTreeNodeDebug(struct tree *T, int level, double *x)
{
  int i,k;
  printf("\nLevel = %d, nObs = %d, left interval = %f, right interval = %f\n",
	 level,T->nObs,
	 T->intervalLeft,T->intervalRight);
  printf("\tmaxPolyOrder = %d, optPolyOrder = %d\n",
	 T->maxPolyOrder,T->optPolyOrder);
  for (i = 0; i < T->maxPolyOrder; i++) {
    printf("%d order poly coeffs:\t",i+1);
    for (k = 0; k <= i; k++) {
      printf("\t%lf",T->polyCoeffs[i][k]);
    }
    printf("\n");
  }
  printf("Likelihoods\n");
  for (i = 0; i < T->maxPolyOrder; i++) {
    printf("\t%lf",T->likelihoods[i]);
  }
  printf("\n");
  /*
  for (i = 0; i < T->nObs; i++)
  {
    printf("\t%lf",x[i+T->firstObsInInterval]);
  }
  if (i > 0)
    printf("\n");
  */
}
void printTreeDebug(struct tree *T, int level, double *x)
{
  if (T == NULL) 
    return;
  printTreeNodeDebug(T,level,x);
  printTreeDebug(T->leftChild,level+1,x);
  printTreeDebug(T->rightChild,level+1,x);
}

void freeTree(struct tree *T)
{
	int i;
	/*printf("fT leftChild\n");*/
  if (T->leftChild != NULL)   {
    freeTree(T->leftChild);
  }
	/*printf("fT rightChild\n");*/
  if (T->rightChild != NULL)   {
    freeTree(T->rightChild);
  }
	/*printf("fT polyCoeffs\n");*/
	if (T->polyCoeffs != NULL) {
		/*printf("fT maxPolyOrder = %d\n",T->maxPolyOrder);*/
		for (i = 0; i < T->maxPolyOrder; i++) {
			free(T->polyCoeffs[i]);
		}
		free(T->polyCoeffs);
	} 
	/*printf("fT likelihoods\n");*/
	if (T->likelihoods != NULL) {
		free(T->likelihoods);
	}
	/*printf("fT T\n");*/
  free(T);
}


void coeffsToEval2(double *x, struct tree *T, int polyOrder, double *est) {
  double *y, *da, *db, *dc;
  int i,k;
  int estLength = T->nObs;
  double a,b;
  
  a = T->intervalLeft;
  b = T->intervalRight;
  y = (double *)malloc(sizeof(double)*estLength);
  da = (double *)malloc(sizeof(double)*estLength);
  db = (double *)malloc(sizeof(double)*estLength);
  dc = (double *)malloc(sizeof(double)*estLength);
  for (i = 0; i < T->nObs; i++) {
    y[i] = (x[i+T->firstObsInInterval]-a)/(b-a)*2.0-1.0;
    da[i] = 0.0;
    db[i] = 0.0;
    dc[i] = 0.0;
  }
  for (k = polyOrder-1; k>=1; k--) {
    for (i = 0; i < T->nObs; i++) {
      dc[i] = 2*y[i]*db[i]-da[i]+T->polyCoeffs[polyOrder-1][k];
      da[i] = db[i];
      db[i] = dc[i];
    }
  }
  for (i = 0; i < T->nObs; i++) {
    est[i] = y[i]*dc[i]-da[i]+T->polyCoeffs[polyOrder-1][0];
  }
  free(y);
  free(da);
  free(db);
  free(dc);
}

void coeffsToEval(double *x, struct tree *T, int polyOrder, double *est) {
  double *cheb, *chebPrev, *chebNext, *y;
  int i,k;
  int estLength = T->nObs;
  double a,b;
  
  a = T->intervalLeft;
  b = T->intervalRight;

  y = (double *)malloc(sizeof(double)*estLength);
  cheb = (double *)malloc(sizeof(double)*estLength);
  chebNext = (double *)malloc(sizeof(double)*estLength);
  chebPrev = (double *)malloc(sizeof(double)*estLength);
  for (i = 0; i < T->nObs; i++) {
    y[i] = (x[i+T->firstObsInInterval]-a)/(b-a)*2.0-1.0;
    cheb[i] = 1.0;
    est[i] = 0.0;
  }
  for (k = 0; k < polyOrder; k++) {
    for (i = 0; i < estLength; i++) {
      est[i] += cheb[i]*T->polyCoeffs[polyOrder-1][k];
    }
    if (k==0) {
      for (i = 0; i < estLength; i++) {
	chebNext[i] = y[i];
      }
    }
    else {
      for (i = 0; i < estLength; i++) {
	chebNext[i] = 2.0*y[i]*cheb[i] - chebPrev[i];
      }
    }
    
    for (i = 0; i < estLength; i++) {
      chebPrev[i] = cheb[i];
      cheb[i] = chebNext[i];
    }
  }
  free(y);
  free(cheb);
  free(chebPrev);
  free(chebNext);
}

void coeffsToEst2(int estLength, int estStart, double *coeffs, 
		 int numCoeffs, double *est) {
  double *y, *da, *db, *dc;
  int i,k;
  
  y = (double *)malloc(sizeof(double)*estLength);
  da = (double *)malloc(sizeof(double)*estLength);
  db = (double *)malloc(sizeof(double)*estLength);
  dc = (double *)malloc(sizeof(double)*estLength);
  for (i = 0; i < estLength; i++) {
    y[i] = (double)i/(double)estLength*2.0-1.0;
    da[i] = 0.0;
    db[i] = 0.0;
    dc[i] = 0.0;
  }
  for (k = numCoeffs-1; k>=1; k--) {
    for (i = 0; i < estLength; i++) {
      dc[i] = 2*y[i]*db[i]-da[i]+coeffs[k];
      da[i] = db[i];
      db[i] = dc[i];
    }
  }
  for (i = 0; i < estLength; i++) {
    est[i+estStart] = y[i]*dc[i]-da[i]+coeffs[0];
  }
  free(y);
  free(da);
  free(db);
  free(dc);
}

void coeffsToEst(int estLength, int estStart, double *coeffs, 
		 int numCoeffs, double *est) {
  double *cheb, *chebPrev, *chebNext, *y;
  int i,k;

  cheb = (double *)malloc(sizeof(double)*estLength);
  chebNext = (double *)malloc(sizeof(double)*estLength);
  chebPrev = (double *)malloc(sizeof(double)*estLength);
  y = (double *)malloc(sizeof(double)*estLength);

  for (i = 0; i < estLength; i++) {
    cheb[i] = 1.0;
    est[i+estStart] = 0.0;
    y[i] = (double)i/(double)estLength*2.0-1.0;
  }
  for (k = 0; k < numCoeffs; k++) {
    for (i = 0; i < estLength; i++) {
      est[i+estStart] += cheb[i]*coeffs[k];
    }
    if (k==0) {
      for (i = 0; i < estLength; i++) {
	chebNext[i] = y[i];
      }
    }
    else {
      for (i = 0; i < estLength; i++) {
	chebNext[i] = 2*y[i]*cheb[i] - chebPrev[i];
      }
    }
    
    for (i = 0; i < estLength; i++) {
      chebPrev[i] = cheb[i];
      cheb[i] = chebNext[i];
    }
  }
  free(y);
  free(cheb);
  free(chebPrev);
  free(chebNext);
}

void treeToEst(struct tree *T, double *est, 
	       int estLength, int estStart, int level)
{
  int breakpoint;
  
  if (T->optPolyOrder == -1){
    /* split */
    breakpoint = (int) floor(estLength/2);
    treeToEst(T->leftChild, est, breakpoint, estStart,level+1);
    treeToEst(T->rightChild, est, estLength-breakpoint, 
	      estStart+breakpoint,level+1);
  }
  else {
    /* prune */
    coeffsToEst(estLength,estStart,T->polyCoeffs[T->optPolyOrder-1],
		T->optPolyOrder,est);
  }
}
